#
# Debugging:
#
# To use the pdb debugger on code in this file, you will need to disable
# the parts of NUU that capture output for the Log window.  Do that in
# two places.  In YumIO.c, change:
#    #undef DEBUG_DISABLE_CAPTURERS
#       =>
#    #define DEBUG_DISABLE_CAPTURERS
# Then just insert the following wherever you need to gain control with pdb:
#    debug()
# If stdin, stdout & stderr still give you troubles, you also could change this
# at the end of this file:
#    #sys.stdout = YumIOStdout()
#    #sys.stderr = YumIOStderr()
#    #sys.stdin = YumIOStdin()
#

import sys
import os
sys.path.insert(0, '%s/python' % os.getenv('NUUSHARE', '/usr/share/nuu'))

from pprint import pprint

def debug():
    import sys
    old_stdout = sys.stdout
    sys.stdout = sys.__stdout__
    import pdb
    pdb.set_trace()
    sys.stdout = old_stdout

# To see libdnf & dnf debug logging, add '-v' to this option set.
# YumIO_dnf_options = [ 'makecache', '-v' ]
YumIO_dnf_options = [ 'makecache' ]

# If the garbage collector is enabled (the default), if anyone does an
# update, presses Cancel, does it again, and presses OK, the update will
# complete successfully, but then the re-initialization will fail with an
# assertion:
#    Modules/gcmodule.c:378: delete_garbage:
#    Assertion `((((PyGC_Head *)(op)-1))->gc.gc_refs >= 0)' failed.
# I went so far as disabling every single Py_DECREF in the C++ code and
# verified that there were Py_INCREF's before the calls to PyTuple_SetItem
# and PyList_SetItem, just to make sure I wasn't leaving any dangling
# references, but the gc still blew up.  I suspect this is a bug in Python's
# gc, so I'm disabling it.
import gc
gc.disable()

# For versions of Python that do GIL locking, NUU must act as a multi-threaded
# application (even though NUU itself guarantees that python cannot be running
# in more than one thread at a time).  But the YumBaseCli.doTransaction will
# attempt to call signal.signal, and that is not allowed in any non-main thread.
# To workaround this, create a new YumIOSignal that mostly works like the
# original signal module, but where the signal function is overridden to do no
# real work.
import types
import signal
__signal__ = signal
YumIOSignal = types.ModuleType('YumIOSignal', 'YumIO replacement for signal')
for meth in dir(__signal__):
    if meth not in [ '__name__', '__doc__', '__file__' ]:
        setattr(YumIOSignal, meth, getattr(__signal__, meth))
YumIOSignal.signal = lambda num, obj: __signal__.getsignal(num)
sys.modules['signal'] = YumIOSignal
import signal

#
# Main body of YumIO
#

import locale
import dnf.cli
import dnf.cli.commands.clean
import yumIOC
import dnf.yum
import dnf.rpm
import rpm
import distutils.version
import copy

def dnfVersion():
    if hasattr(dnf, '__version__'):
        return distutils.version.StrictVersion(dnf.yum.__version__)
    else:
        return distutils.version.StrictVersion('0.0')

YumIO_group_pkg_types = ('mandatory', 'default', 'conditional')

# Used to override sys.stdin so that it will claim that it isatty, which
# convinces a yum check that it is not running unattended.
__stdin__ = sys.stdin
class YumIOStdin:
    def __init__(self):
        pass
    def isatty(self):
        return 1
for meth in dir(__stdin__):
    if not hasattr(YumIOStdin, meth):
        setattr(YumIOStdin, meth, getattr(__stdin__, meth))

# Used to override sys.stdout
class YumIOStdout:
    def write(self, msg):
        yumIOC.stdoutWrite(msg)

    def isatty(self):
        return 0

    def flush(self):
        pass

# Used to override sys.stderr
class YumIOStderr:
    def write(self, msg):
        yumIOC.stderrWrite(msg)

    def isatty(self):
        return 0

    def flush(self):
        pass

    # rpmtrans calls rpm.setLogFile(sys.stderr), and that demands a stderr with
    # a working PyObject_AsFileDescriptor.  So, provide one.
    def fileno(self):
        return sys.__stderr__.fileno()

# Use to override Cli's stdout and stderr handlers
import logging
class YumIOHandler(logging.Handler):
    def __init__(self, name = None):
        super(YumIOHandler, self).__init__()
    
    def emit(self, record):
        # The numeric values correspond to the ascending levels of detail in
        # LogWindow's "Detail level".  The values are arbitrary.  They need to
        # match the code in Log::DetailSpinBox::textFromValue.
        if record.message.startswith("Cannot download "):
            # This is a DEBUG (at least in dnf-4.2), but it should be an ERROR.
            # Elevate it.
            return yumIOC.errorlog(0, record.message)
        elif record.levelno >= dnf.logging.ERROR:
            return yumIOC.errorlog(0, record.message)
        elif record.levelno >= dnf.logging.WARNING:
            return yumIOC.warnlog(1, record.message)
        elif record.levelno >= dnf.logging.INFO:
            return yumIOC.log(2, record.message)
        elif record.levelno >= dnf.logging.DEBUG:
            return yumIOC.log(3, record.message)
        elif record.levelno >= dnf.logging.DDEBUG:
            return yumIOC.log(4, record.message)
        elif record.levelno >= dnf.logging.SUBDEBUG:
            return yumIOC.log(5, record.message)
        else:
            return yumIOC.log(6, record.message)
    
# Used to override BaseCli's userconfirm & transaction callbacks
class YumIOOutput(dnf.cli.output.Output):
    def userconfirm(self, msg=None, defaultyes_msg=None):
        sys.stdout.flush()
        sys.stderr.flush()
        return yumIOC.userconfirm()

    def reportDownloadSize(self, packages, install_only=False):
        yumIOC.reportDownloadSize(packages)
        return super(YumIOOutput, self).reportDownloadSize(packages,
                                                           install_only)

    def reportRemoveSize(self, packages):
        yumIOC.reportRemoveSize(packages)
        return super(YumIOOutput, self).reportRemoveSize(packages)

    def list_transaction(self, transaction):
        yumIOC.listTransaction(self, transaction)
        return super(YumIOOutput, self).list_transaction(transaction)

    def post_transaction_output(self, transaction):
        yumIOC.postTransactionOutput(self, transaction)
        return super(YumIOOutput, self).post_transaction_output(transaction)

# Used to replace the dnf.cli.progress.MultiFileProgressMeter, overriding each
# repo's progress bar (via set_progress_bar).
class YumIOMultiFileProgressMeter(dnf.callback.DownloadProgress):
    def start(self, total_files, total_size, total_drpms=0):
        return yumIOC.downloadStart(total_files, total_size, total_drpms)

    def progress(self, payload, done):
        return yumIOC.downloadProgress(payload, done)

    def message(self, msg):
        return yumIOC.downloadMessage(total_files, total_size)

    def end(self, payload, status, msg):
        return yumIOC.downloadEnd(payload, status, msg)

# YumIOTransactionDisplay replaces the dnf.cli.output.CliTransactionDisplay,
# which is passed as part of the display list in dnf.cli.BaseCli.do_transaction.
#
# However, that do_transaction does a LOT and it would be problematic to
# replicate it.  So I need to trick it into creating my own TransactionDisplay
# when it wants to create its own dnf.cli.output.CliTransactionDisplay.  Do this
# by breaking into the dnf.cli.output module and replacing the class there with
# my own.
class YumIOTransactionDisplay(dnf.yum.rpmtrans.LoggingTransactionDisplay):
    def __init__(self):
        super(YumIOTransactionDisplay, self).__init__()

    def progress(self, package, action, ti_done, ti_total, ts_done, ts_total):
        return yumIOC.rpmtransProgress(package, action,
                                       ti_done, ti_total,
                                       ts_done, ts_total)

    def scriptout(self, msgs):
        return yumIOC.rpmtransScriptout(msgs)
dnf.cli.output.CliTransactionDisplay = YumIOTransactionDisplay
        
# YumIORPMTransaction replaces the dnf.yum.rpmtrans.RPMTransaction, which is
# created twice in dnf.Base.do_transaction, once for the transaction test
# (a.k.a. "test"), and again for the real transaction (a.k.a. "run").
#
# However, that do_transaction does a LOT and it would be problematic to
# replicate it.  So I need to trick it into creating my own RPMTransaction when
# it wants to create its own dnf.yum.rpmtrans.RPMTransaction.  Do this by
# breaking into the dnf.yum.rpmtrans module and replacing the class there with
# my own.
#
# This class goes beyond the callbacks defined in the TransactionDisplay above
# by also making callbacks when starting and ending a transaction test or
# transaction run.  It detects that by intercepting the __init__ and __del__
# methods.  Alternatively, it could've intercepted the callback method and
# checked the "what" for RPMCALLBACK_TRANS_START or RPMCALLBACK_TRANS_STOP, but
# that's more complicated and more likely to break because of later dnf changes.
class YumIORPMTransaction(dnf.yum.rpmtrans.RPMTransaction):
    def __init__(self, base, test=False, displays=()):
        yumIOC.rpmtransInit(test)
        super(YumIORPMTransaction, self).__init__(base,
                                                  test=test, displays=displays)

    def __del__(self):
        yumIOC.rpmtransDel()
        super(YumIORPMTransaction, self).__del__()
dnf.yum.rpmtrans.RPMTransaction = YumIORPMTransaction

class YumBaseCli(dnf.cli.cli.BaseCli):
    def _download_remote_payloads(self, payloads, drpm, progress,
                                  callback_total):
        yumIOC.downloadRemotePayloads(payloads);
        return super(YumBaseCli, self)._download_remote_payloads(payloads, drpm,
                                                                 progress,
                                                                 callback_total)

    def gpgsigcheck(self, pkgs):
        yumIOC.gpgsigcheck()
        return super(YumBaseCli, self).gpgsigcheck(pkgs)

# This is the minimum necessary argparse.Namespace object with fake options
# to mimic what "dnf clean" does.  It intentionally omits the type attribute,
# because each user will need to set that to the desired values.
from argparse import Namespace
YumIO_default_clean_options = Namespace(command=['clean'])

class YumIO(dnf.cli.cli.Cli):
    def __init__(self):
        locale.setlocale(locale.LC_ALL, '')
        basecli = YumBaseCli(dnf.conf.Conf())
        super(YumIO, self).__init__(basecli)
        super(YumIO, self).configure(YumIO_dnf_options)

        # Override the standard dnf logging stdout & stderr StreamHandlers
        # with a single custom YumIOHandler.
        logger_dnf = logging.getLogger("dnf")
        logger_dnf.removeHandler(self.base._logging.stdout_handler)
        logger_dnf.removeHandler(self.base._logging.stderr_handler)
        self.base._logging.stdout_handler = None
        self.base._logging.stderr_handler = None
        logger_dnf.addHandler(YumIOHandler())

        self.base.output = YumIOOutput(self.base, self.base.conf)
        self.initOrReset()

    # Utility functions

    def initOrReset(self):
        bar = YumIOMultiFileProgressMeter()
        self.base.repos.all().set_progress_bar(bar)
        self.base.output.progress = bar

    def log(self, value, msg):
        return yumIOC.log(value, msg)

    def errorlog(self, value, msg):
        return yumIOC.errorlog(value, msg)

    def lock(self):
        pass

    def unlock(self):
        pass

    def enableRepo(self, id):
        repo = self.base.repos.get_matching(id)
        repo.enable()

    def disableRepo(self, id):
        repo = self.base.repos.get_matching(id)
        repo.disable()
    
    # High-level functions called from YumIO.c

    def setupRepos(self):
        try:
            self.log(2, 'Setting up Repos')
            # It proved impossible to load 'otherdata' metadata on demand after
            # initialization.  So instead, ensure it is loaded during
            # initialization regardless of whether or not it will be needed.
            for repo in self.base.repos.iter_enabled():
                repo.load_metadata_other = True
            self.base.fill_sack(load_system_repo='auto',
                                load_available_repos=True)
            return None
        except:
            raise

    def collectUpdates(self):
        return self.base.returnPkgLists('upgrades').updates

    def collectAll(self):
        return self.base.returnPkgLists('all')

    def collectGroups(self):
        self.base.read_comps(arch_filter=True)
        return self.base.comps.groups

    def collectDone(self):
        self.base._closeRpmDB()

    def apply(self,
              installs, updates, removes,
              grp_installs, grp_updates, grp_removes):
        try:
            for grp in grp_installs:
                self.base.group_install(grp.id, YumIO_group_pkg_types)
            for grp in grp_updates:
                # NOTE: OUT_OF_DATE groups actually are considered "available"
                # (a.k.a. uninstalled) by dnf.  So, to update them, really just
                # "install" them.
                self.base.group_install(grp.id, YumIO_group_pkg_types)
            for grp in grp_removes:
                self.base.group_remove(grp.id)
            for pkg in installs:
                self.base.package_install(pkg)
            for pair in updates:
                pkg_spec = pair[0].name + '-' + pair[0].evr + '.' + pair[0].arch
                self.base.upgrade(pkg_spec)
            for pkg in removes:
                self.base.package_remove(pkg)
            result = self.base.resolve()
            if not result:
                raise Error.YumBaseError('Nothing to do')
            self.base.do_transaction()
            # If the transaction succeeds, we need to re-load the sack.
            self.base.reset(goal=True, sack=True)
        except:
            self.base.reset(goal=True)
            if hasattr(self, 'history'):
                self.history.close()
            self.base._closeRpmDB()
            self.initOrReset()
            raise

    def collectFileLists(self):
        # This is loaded all the time by default by fill_sack.
        pass

    def collectOtherData(self):
        # It was not possible to load this lazily, so instead setupRepos sets
        # the load_metadata_other flags, causing fill_sack to load it on startup.
        pass

    def cleanType(self, type):
        cmd = dnf.cli.commands.clean.CleanCommand(self)
        cmd.opts = copy.copy(YumIO_default_clean_options)
        cmd.opts.type = type
        return cmd.run()

    def cleanDbcache(self):
        return self.cleanType([ 'dbcache' ])

    def cleanPackages(self):
        return self.cleanType([ 'packages' ])

    def cleanMetadata(self):
        return self.cleanType([ 'metadata' ])

    def cleanAll(self):
        return self.cleanType([ 'all' ])

    def reset(self):
        self.base.reset(goal=True, repos=True, sack=True)
        if hasattr(self, 'history'):
            self.history.close()
        self.base._closeRpmDB()
        # Re-create the nascent set of repos.
        # They'll be re-read during the next collect.
        opts = self.optparser.parse_main_args(YumIO_dnf_options)
        super(YumIO, self)._configure_repos(opts)
        self.initOrReset()
        
    def getBiArch(self, arch):
        try:
            return dnf.rpm.basearch(arch)
        except KeyError:
            return arch

    def isBetterThan(self, pkg1, pkg2):
        if self.getBiArch(pkg1.arch) == self.base.conf.basearch:
            return True
        elif self.getBiArch(pkg2.arch) == self.base.conf.basearch:
            return False
        else:
            return True

sys.stdout = YumIOStdout()
sys.stderr = YumIOStderr()
sys.stdin = YumIOStdin()
